import template from '@babel/template';
import * as types from '@babel/types';

const isReactiveStatement = (path = {}) =>
    path.node &&
    path.node.label &&
    types.isIdentifier(path.node.label, { name: '$' });

const getInducerSubscriptionStatement = ({ inducers = [] }) => {
    const buildAst = template(`
        %%subscribeFunctionCall%%
    `);

    return inducers.map(inducer => {
        const identifier = types.identifier(inducer.subscriptionFunctionName);
        const subscribeFunctionCall = types.callExpression(identifier, []);
        return buildAst({
            subscribeFunctionCall
        });
    });
};

const transformReactiveAssignmentExpression = ({ scope, statement }) => {
    const left = statement.expression.left;
    const evalFunctionIdentifier = scope.generateUidIdentifier(
        `${left.name}$eval`
    );
    const evalFunctionCall = types.callExpression(evalFunctionIdentifier, []);

    const buildAst = template(`
            const %%evalFunctionIdentifier%% = () => {
                %%statement%%
            };
            %%evalFunctionCall%%
        `);
    return buildAst({
        evalFunctionIdentifier,
        statement,
        evalFunctionCall
    });
};

const transformInducerDeclaration = ({ name, scope, statement }) => {
    const subscribeFunctionIdentifier = scope.generateUidIdentifier(
        `${name}$sub`
    );

    const buildAst = template(`
        %%statement%%
        const %%subscribeFunctionIdentifier%% = (() => {
            const subscribers = [];
            const subscribe = subscriber => {
                subscribers.push(subscriber);
            };
            return subscribe;
        })();
    `);

    const statements = buildAst({
        statement,
        subscribeFunctionIdentifier
    });

    return {
        name: subscribeFunctionIdentifier.name,
        statements
    };
};

const inducerVisitor = {
    Identifier(path, state) {
        const name = path.node.name;
        if (!path.scope.hasBinding(name)) {
            path.skip();
            return;
        }

        const inducerBinding = path.scope.getBinding(name);
        const inducerBindingPath = inducerBinding.path;
        if (!inducerBindingPath.isVariableDeclarator()) {
            path.skip();
            return;
        }

        const inducerDeclarationPath = inducerBindingPath.parentPath;
        const {
            name: subscriptionFunctionName,
            statements
        } = transformInducerDeclaration({
            name,
            scope: inducerDeclarationPath.scope,
            statement: inducerDeclarationPath.node
        });
        inducerDeclarationPath.replaceWithMultiple(statements);

        const reactiveStatement =
            state.reactiveStatements[state.reactiveStatements.length - 1];
        reactiveStatement.inducers.push({ name, subscriptionFunctionName });
    }
};

const ReactiveStatement = {
    LabeledStatement: {
        enter(path, state) {
            if (!isReactiveStatement(path)) {
                path.skip();
                return;
            }

            const statement = path.node.body;
            if (!types.isExpressionStatement(statement)) {
                path.skip();
                return;
            }

            const expression = statement.expression;
            if (!types.isAssignmentExpression(expression)) {
                path.skip();
                return;
            }

            if (!state.reactiveStatements) {
                state.reactiveStatements = [];
            }

            const reactiveStatementState = {
                filter: [],
                inducers: []
            };
            state.reactiveStatements.push(reactiveStatementState);
            const expressionRightPath = path.get('body.expression.right');
            expressionRightPath.traverse(inducerVisitor, state);
        },
        exit(path, state) {
            const reactiveStatementState = state.reactiveStatements.pop();
            const inducers = reactiveStatementState.inducers || [];
            const statement = path.node.body;
            let statements = transformReactiveAssignmentExpression({
                scope: path.scope,
                statement
            });
            statements = statements.concat(
                getInducerSubscriptionStatement({ inducers })
            );
            path.replaceWithMultiple(statements);
        }
    }
};

export default ReactiveStatement;
