// Libraries
import { parse } from '@babel/parser';
import traverse from '@babel/traverse';
import generate from '@babel/generator';
// Visitors
import ReactiveStatementVisitor from './index';

const compile = code => {
    const ast = parse(code);
    traverse(ast, ReactiveStatementVisitor, null, {});
    return generate(ast).code;
};

describe('Reactive statement visitor', () => {
    describe('Assignment expressions', () => {
        it('transpiles a reactive statement', () => {
            const transpiledCode = compile(`
                let a = 2;
                let b = 4;
                let test;
                $: test = a * b;
            `);
            expect(transpiledCode).toMatchInlineSnapshot(`
                "let a = 2;

                const _a$sub = (() => {
                  const subscribers = [];

                  const subscribe = subscriber => {
                    subscribers.push(subscriber);
                  };

                  return subscribe;
                })();

                let b = 4;

                const _b$sub = (() => {
                  const subscribers = [];

                  const subscribe = subscriber => {
                    subscribers.push(subscriber);
                  };

                  return subscribe;
                })();

                let test;

                const _test$eval = () => {
                  test = a * b;
                };

                _test$eval();

                _a$sub();

                _b$sub();"
            `);
        });

        it('transpiles a reactive statement in non global scope', () => {
            const transpiledCode = compile(`
                const test = () => {
                    let count = 1;
                    let doubled;
                    $: doubled = count * 2;
                };
            `);
            expect(transpiledCode).toMatchInlineSnapshot(`
                "const test = () => {
                  let count = 1;

                  const _count$sub = (() => {
                    const subscribers = [];

                    const subscribe = subscriber => {
                      subscribers.push(subscriber);
                    };

                    return subscribe;
                  })();

                  let doubled;

                  const _doubled$eval = () => {
                    doubled = count * 2;
                  };

                  _doubled$eval();

                  _count$sub();
                };"
            `);
        });

        xit('transpiles multiple reactive statements', () => {
            const transpiledCode = compile(`
                let count = 1;
                let doubled;
                let tripled;
                $: doubled = count * 2;
                $: tripled = count * 2;
            `);

            expect(transpiledCode).toMatchInlineSnapshot(`
                "let count = 1;
                let doubled;
                let tripled;

                const _doubled$eval = () => {
                  doubled = count * 2;
                };

                _doubled$eval();

                _count$sub && _count$sub();

                const _tripled$eval = () => {
                  tripled = count * 2;
                };

                _tripled$eval();

                _count$sub2 && _count$sub2();"
            `);
        });

        it('avoids colliding with identifier', () => {
            const transpiledCode = compile(`
                let _doubled$eval;
                let count = 1;
                let doubled;
                $: doubled = count * 2;
            `);
            expect(transpiledCode).toMatchInlineSnapshot(`
                "let _doubled$eval;

                let count = 1;

                const _count$sub = (() => {
                  const subscribers = [];

                  const subscribe = subscriber => {
                    subscribers.push(subscriber);
                  };

                  return subscribe;
                })();

                let doubled;

                const _doubled$eval2 = () => {
                  doubled = count * 2;
                };

                _doubled$eval2();

                _count$sub();"
            `);
        });

        it('avoids colliding with identifier in outer scope', () => {
            const transpiledCode = compile(`
                let _doubled$eval;
                const test = () => {
                    let count = 1;
                    let doubled;
                    $: doubled = count * 2;
                };
            `);
            expect(transpiledCode).toMatchInlineSnapshot(`
                "let _doubled$eval;

                const test = () => {
                  let count = 1;

                  const _count$sub = (() => {
                    const subscribers = [];

                    const subscribe = subscriber => {
                      subscribers.push(subscriber);
                    };

                    return subscribe;
                  })();

                  let doubled;

                  const _doubled$eval2 = () => {
                    doubled = count * 2;
                  };

                  _doubled$eval2();

                  _count$sub();
                };"
            `);
        });

        it("avoid inducers that aren't binded", () => {
            const transpiledCode = compile(`
                let b = 2;
                let test;
                $: test = a * b;
            `);
            expect(transpiledCode).toMatchInlineSnapshot(`
                "let b = 2;

                const _b$sub = (() => {
                  const subscribers = [];

                  const subscribe = subscriber => {
                    subscribers.push(subscriber);
                  };

                  return subscribe;
                })();

                let test;

                const _test$eval = () => {
                  test = a * b;
                };

                _test$eval();

                _b$sub();"
            `);
        });

        it("doesn't transpile non-reactive statements", () => {
            const transpiledCode = compile('let a = 1;');
            expect(transpiledCode).toBe('let a = 1;');
        });

        it("doesn't transpile labeled statements with non-reactive labels", () => {
            const transpiledCode = compile('test: a = 1;');
            expect(transpiledCode).toBe('test: a = 1;');
        });
    });
});
